#ifdef _WIN32
#  include "winconfig.h"
#  define XMLPARSEAPI(type) __declspec(dllexport) type __cdecl
#  include "expat.h"
#  undef XMLPARSEAPI
#else
#include "config.h"

#ifdef __declspec
#  define XMLPARSEAPI(type) __declspec(dllexport) type __cdecl
#endif

#include "expat.h"

#ifdef __declspec
#  undef XMLPARSEAPI
#endif
#endif /* ndef COMPILED_FROM_DSP */

#include <stddef.h>
#include <string.h>

#ifdef XML_UNICODE
#define XML_ENCODE_MAX XML_UTF16_ENCODE_MAX
#define XmlConvert XmlUtf16Convert
#define XmlGetInternalEncoding XmlGetUtf16InternalEncoding
#define XmlGetInternalEncodingNS XmlGetUtf16InternalEncodingNS
#define XmlEncode XmlUtf16Encode
#define MUST_CONVERT(enc, s) (!(enc)->isUtf16 || (((unsigned long)s) & 1))
typedef unsigned short ICHAR;
#else
#define XML_ENCODE_MAX XML_UTF8_ENCODE_MAX
#define XmlConvert XmlUtf8Convert
#define XmlGetInternalEncoding XmlGetUtf8InternalEncoding
#define XmlGetInternalEncodingNS XmlGetUtf8InternalEncodingNS
#define XmlEncode XmlUtf8Encode
#define MUST_CONVERT(enc, s) (!(enc)->isUtf8)
typedef char ICHAR;
#endif


#ifndef XML_NS

#define XmlInitEncodingNS XmlInitEncoding
#define XmlInitUnknownEncodingNS XmlInitUnknownEncoding
#undef XmlGetInternalEncodingNS
#define XmlGetInternalEncodingNS XmlGetInternalEncoding
#define XmlParseXmlDeclNS XmlParseXmlDecl

#endif

#ifdef XML_UNICODE_WCHAR_T
#define XML_T(x) L ## x
#else
#define XML_T(x) x
#endif

#include "expatll.h"

const ENCODING *enc;
const ENCODING *internalEncoding;
INIT_ENCODING initEnc;
ATTRIBUTE* attributes;
static int attributesSize;
static int* next = NULL;
static int next_dim = 0;


staticforward PyObject* pysearch(PyObject* self, PyObject* args);

static void splitPrefix(char* ptr, int* prefixLen, int* tagLen) {
	int i;
	
	*prefixLen = 0;
	*tagLen = XmlNameLength(enc, ptr);
	for (i=0; i < *tagLen; i++) 
		if (ptr[i] == ':') {
			*prefixLen = i;
			*tagLen -= *prefixLen+1;
			break;
		}
}

static void expatll_parseTag(struct XmlItem_c* item, char* ptr) {
	int tlen,plen;

	ptr++;
	splitPrefix(ptr, &plen, &tlen);
	if (plen != 0) {
		item->prefix = PyString_FromStringAndSize(ptr, plen);
		item->tag =  PyString_FromStringAndSize(ptr+plen+1, tlen);
	}
	else {
		item->prefix = PyString_InternFromString("");	
		item->tag = PyString_FromStringAndSize(ptr, tlen);
	}
}

static void expatll_parseAttrs(struct XmlItem_c* item, char* ptr) {
	char *rawName;
	int tlen,plen;
	int attsSize;
	int i=0;
	PyObject* xattr;
	PyObject* key;
	PyObject* prefix;
	PyObject* value;

	attsSize = XmlGetAttributes(enc, ptr, attributesSize, attributes);
	if (!attsSize) {
		Py_INCREF(Py_None);
		item->attrs = Py_None;
		return;
	}
	
	if (attsSize > attributesSize) {
		attributesSize = attsSize+1;
		free(attributes);
		attributes = malloc(sizeof(ATTRIBUTE)*attributesSize);
		attsSize = XmlGetAttributes(enc, ptr, attributesSize, attributes);
	}

	item->attrs = PyList_New(0);
	for (i=0; i < attsSize; ++i) {
		rawName = (char*) attributes[i].name;

		splitPrefix(rawName, &plen, &tlen);
		if (plen != 0) {
			prefix = PyString_FromStringAndSize(rawName, plen);
			key = PyString_FromStringAndSize(rawName+plen+1, tlen);
		}
		else {
			Py_INCREF(Py_None);
			prefix =  Py_None;	
			key = PyString_FromStringAndSize(rawName, tlen);
		}
		
		value = PyString_FromStringAndSize(attributes[i].valuePtr, attributes[i].valueEnd-attributes[i].valuePtr);
		xattr = PyTuple_New(3);
		PyTuple_SET_ITEM(xattr, 0, prefix);
		PyTuple_SET_ITEM(xattr, 1, key);
		PyTuple_SET_ITEM(xattr, 2, value);
		PyList_Append(item->attrs, (PyObject*) xattr);
		Py_DECREF(xattr);
	}
}

static void expatll_parseStartTag(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_startTag;
	item->len = *next-ptr;

	expatll_parseTag(item, ptr);
	expatll_parseAttrs(item, ptr);
}

static void expatll_parseEmptyElement(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_emptyElement;
	item->len = *next-ptr;

	expatll_parseTag(item, ptr);
	expatll_parseAttrs(item, ptr);
}

static void expatll_parseChar(struct XmlItem_c* item, char value) {
	item->type = XmlItem_data; 	
	item->len = 1;
	item->value = PyString_FromStringAndSize(&value, 1);
}

static void expatll_parseDataChars(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_data; 	
	item->len = *next-ptr;
	item->value = PyString_FromStringAndSize(ptr, *next-ptr);
}

static void expatll_parseComment(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_comment; 	
	item->len = *next-ptr-7;
	item->value = PyString_FromStringAndSize(ptr+4, item->len);
}

static void expatll_parseXmlDecl(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_XmlDecl;
	item->len = *next-ptr;

	expatll_parseTag(item, ptr);
	expatll_parseAttrs(item, ptr);
}


static void expatll_parsePI(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_PI;
	item->len = *next-ptr;

	expatll_parseTag(item, ptr);
	expatll_parseAttrs(item, ptr);
}

static void expatll_parseCDATA(struct XmlItem_c* item, char* ptr, char** next) {
	expatll_parseDataChars(item, ptr, next);
	item->type = XmlItem_cdata;
}


static void expatll_parseEndTag(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_endTag;
	item->len = *next-ptr;
}

static void expatll_parseUnknown(struct XmlItem_c* item, char* ptr, char** next) {
	item->type = XmlItem_unknown;
	item->len = *next-ptr;
}

void expatll_parse(struct XmlItem_c* item, char* start, char* end, char** next) {
	int code = XmlContentTok(enc, (const char*)start, end, (const char**)next);

	if (!strncmp(start, "</>", 3)) {
		item->type = XmlItem_endTag;
		item->len = 3;
		*next = start+3; 
		return;
	}

	switch(code) {
		case XML_TOK_PARTIAL: item->type = XmlItem_partial; return;
		case XML_TOK_START_TAG_WITH_ATTS: expatll_parseStartTag(item, start, next); break;
		case XML_TOK_START_TAG_NO_ATTS: expatll_parseStartTag(item, start, next); break;
		case XML_TOK_EMPTY_ELEMENT_WITH_ATTS: expatll_parseEmptyElement(item, start, next); break;
		case XML_TOK_EMPTY_ELEMENT_NO_ATTS: expatll_parseEmptyElement(item, start, next); break;
		case XML_TOK_END_TAG: expatll_parseEndTag(item, start, next); break;
		case XML_TOK_DATA_CHARS: expatll_parseDataChars(item, start, next); break;
		case XML_TOK_CDATA_SECT_OPEN: expatll_parseCDATA(item, start, next); break;
		case XML_TOK_CHAR_REF: expatll_parseUnknown(item, start,next);  break;
		case XML_TOK_TRAILING_CR: *next = end; break;
		case XML_TOK_COMMENT: expatll_parseComment(item, start, next); break;
		case XML_TOK_PI: expatll_parsePI(item, start, next); break;
		case XML_TOK_XML_DECL: expatll_parseXmlDecl(item, start, next); break;
		case XML_TOK_ENTITY_REF: 
					if (!strncmp(start+1, "gt;", 3))
						expatll_parseChar(item, '>');
					else if (!strncmp(start+1, "lt;", 3))
						expatll_parseChar(item, '<');
					else if (!strncmp(start+1, "amp;", 4))
						expatll_parseChar(item, '&');
					else if (!strncmp(start+1, "quote;", 6))
						expatll_parseChar(item, '"');
					else if (!strncmp(start+1, "apos;", 4))
						expatll_parseChar(item, '\'');
					else
						expatll_parseUnknown(item, start,next);
		default: expatll_parseUnknown(item, start,next); 
	}
}
/*
static
void preprocpat(char* pat){ 
	int i, j;
	int len;

// check for size
	len = strlen(pat);
	if (len > next_dim) {
		free(next);
		next_dim = len;
		next = malloc(sizeof(int)*next_dim);
	}
  
// go go 
	i = 0;
    j = next[0] = -1;
    do { 
		if( j==(-1) || pat[i]==pat[j] ) {
			i++;
			j++;
			next[i] = (pat[j]==pat[i]) ? next[j] : j;
		}
		else j = next[j]; 
	}
	while( pat[i] != 0 );
}

static 
char *search(char* text, char* pat, int* next) { 
	int j;

	if( *pat == 0 ) return( text );

	for( j=0; *text; ) {
	  if( j==(-1) || pat[j] == *text ) { 
		  text++; j++; 
		  if( pat[j] == 0 ) return( text-j );
	  }
	  else j = next[j];
	}
	return( NULL );
}


PyObject* searchData(char* buffer, char** next, PyObject* item, int* pos) {
	char* str;
	char* substr;
	char* ptr = buffer;
	char* end = buffer + strlen(buffer);
	char* pattern = PyString_AsString(PyObject_GetAttrString(item, "value"));
	preprocpat( pattern);
	
	while (ptr < end) {
		int code = XmlContentTok(enc, (const char*)ptr, end, (const char**)next);

		if (code == XML_TOK_DATA_CHARS) {
			str = ptr;
			while (XmlContentTok(enc, (const char*)ptr, end, (const char**)next) == XML_TOK_DATA_CHARS) {
				ptr = *next;
				if (ptr >= end)
					break;
			}
			
			substr = search(str, pattern, ptr);
			if (substr) {
				*pos = substr-buffer;
				return PyString_FromStringAndSize(substr, ptr-substr);
			}

		}
		else
			ptr = *next;

	}
	Py_INCREF(Py_None);
	return Py_None;

}


PyObject* searchElement(char* buffer, char** next, PyObject* item, int* pos) {
	char* ptag;
	char* pprefix;
	char* ptr = buffer;
	char* end = buffer + strlen(buffer);
	char* rawName;
	char** aPrefixes;
	char** aTags;
	char** aValues;
	int tlen,plen;
	PyObject* pylist;
	PyObject* pyattr;
	PyObject* pyvalue;

	int n, i, j;
	int attsSize;
	int pattsSize;
	int match;

	ptag = PyString_AsString(PyObject_GetAttrString(item, "tag"));
	pprefix = PyString_AsString(PyObject_GetAttrString(item, "prefix"));

	pylist = PyObject_GetAttrString(item, "value");
	n = PyList_Size(pylist);
	aPrefixes = malloc(sizeof(char*)*n);
	aTags = malloc(sizeof(char*)*n);
	aValues = malloc(sizeof(char*)*n);

	pattsSize = 0;
	for(i=0, j=0; i < n; i++) {
		pyattr = PyList_GetItem(pylist, i);
		pyvalue = PyObject_GetAttrString(pyattr, "value");
		if (pyvalue == Py_None) 
			continue;

		aValues[j] = PyString_AsString(pyvalue);
		aPrefixes[j] = PyString_AsString(PyObject_GetAttrString(pyattr, "prefix"));

		aTags[j] = PyString_AsString(PyObject_GetAttrString(pyattr, "tag"));
		pattsSize++;
	}


	while (ptr < end) {
		int code = XmlContentTok(enc, (const char*)ptr, end, (const char**)next);

		if (code == XML_TOK_START_TAG_WITH_ATTS || code == XML_TOK_START_TAG_NO_ATTS ||
			code == XML_TOK_EMPTY_ELEMENT_WITH_ATTS || code == XML_TOK_EMPTY_ELEMENT_NO_ATTS) {
			splitPrefix(ptr+1, &plen, &tlen);
			if (plen) {
				if (!pprefix) 
					continue;
				if (strncmp(ptr+1,pprefix,plen))
					continue;
				if (strncmp(ptr+plen+2,ptag,tlen))
					continue;
			}
			else {
				if (pprefix) 
					continue;
				if (strncmp(ptr+1,ptag,tlen))
					continue;
			}

			attsSize = XmlGetAttributes(enc, ptr, attributesSize, attributes);
			if (attsSize > attributesSize) {
				attributesSize = attsSize+1;
				free(attributes);
				attributes = malloc(sizeof(ATTRIBUTE)*attributesSize);
				attsSize = XmlGetAttributes(enc, ptr, attributesSize, attributes);
			}

			for (i=0; i < attsSize; ++i) {
				rawName = (char*) attributes[i].name;
				splitPrefix(rawName, &plen, &tlen);
				
				match = 1;
				if (plen) {
					for (j=0; j < pattsSize; ++j) {
						if (!aPrefixes[j])
							continue;
						if (strncmp(rawName,aPrefixes[j],plen))
							continue;
						if (strncmp(rawName+plen+1,aTags[j],plen))
							continue;
						if (strncmp(attributes[i].valuePtr, aValues[j],attributes[i].valueEnd-attributes[i].valuePtr))
							match = 0;
						else
							match = 1;
					}
				}
				else {
					for (j=0; j < pattsSize; ++j) {
						if (aPrefixes[j])
							continue;
						if (strncmp(rawName,aTags[j],tlen))
							continue;
						if (strncmp(attributes[i].valuePtr, aValues[j],attributes[i].valueEnd-attributes[i].valuePtr))
							match = 0;
						else
							match = 1;
					}
				}
				if (!match)
					break;
			}
			
			if (match) {
				*pos = ptr - buffer;
				if (code == XML_TOK_EMPTY_ELEMENT_WITH_ATTS || code == XML_TOK_EMPTY_ELEMENT_NO_ATTS)
					return expatll_parseEmptyElement(ptr,next);
				else
					return expatll_parseStartTag(ptr, next);
			}
		}
		
		ptr = *next;

	}

	free(aPrefixes);
	free(aTags);
	free(aValues);

	Py_INCREF(Py_None);
	return Py_None;

	
}

static PyObject* pysearch(PyObject* self, PyObject* args) {
	char* ptr;
	char* next;
	int len;
	PyObject* pattern;
	PyObject* item;
	PyObject* tuple;
	int tpy;
	int pos;

	if (!PyArg_ParseTuple(args, "siO", &ptr, &len, &pattern))
		return NULL;

	tpy = PyInt_AsLong(PyObject_GetAttrString(pattern, "type"));
	if (tpy == data)
		item = searchData(ptr, &next, pattern, &pos);
	else if (tpy == startTag || tpy == emptyElement)
		item = searchElement(ptr, &next, pattern, &pos);
	else {
		Py_INCREF(Py_None);
		return Py_None;
	}
	
	tuple =  PyTuple_New(2);
	PyTuple_SetItem(tuple, 0, item);
	PyTuple_SetItem(tuple, 1, PyInt_FromLong(pos));
	return tuple;
}

static PyObject* pyparse(PyObject* self, PyObject* args) {
	char* ptr;
	char* next;
	PyObject* pylist;
	int len;

	if (!PyArg_ParseTuple(args, "si", &ptr, &len))
		return NULL;
	
	next = ptr+len;
	pylist = expatll_parse(ptr, &next);
	return pylist;
}
*/
void 
initexpatll(void) 
{
	attributesSize = 20;
	attributes = malloc(sizeof(ATTRIBUTE)*attributesSize);

	next = (int*) malloc(sizeof(int)*20);
	next_dim = 20;

	XmlInitEncoding(&initEnc, &enc, 0);

}
