
/* XList object interface */

/*
Another generally useful object type is an list of object pointers.
This is a mutable type: the list items can be changed, and items can be
added or removed.  Out-of-range indices or non-list objects are ignored.

*** WARNING *** PyList_SetItem does not increment the new item's reference
count, but does decrement the reference count of the item it replaces,
if not nil.  It does *decrement* the reference count if it is *not*
inserted in the list.  Similarly, PyList_GetItem does not increment the
returned item's reference count.
*/

#ifndef Py_XLISTOBJECT_H
#define Py_XLISTOBJECT_H

#include <Python.h>
#include "xlistinfo.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
	PyListObject list;
	PyObject* attrs;
	PyXListInfo* tpinfo;
} PyXListObject;

extern PyObject* Attr_FullName(PyObject* prefix, PyObject* key);
extern int Attr_SplitFullName(PyObject* fullname, PyObject** prefix, PyObject** key);

extern void initxlist();
extern PyObject* PyXList_PoorCopy(PyObject* self);
extern PyObject* PyXList_Copy(PyObject* self);

extern PyObject* PyXList_New();
extern DL_IMPORT(PyTypeObject) PyXList_Type;
extern int xlist_match_c(PyXListObject * list, PyXListObject * pattern);
extern PyObject *
xlist_repr_c(PyXListObject *v, PyObject* suggested_prefixes, PyObject* declared_ns, int strict);

#define PyXList_Check(op) PyObject_TypeCheck(op, &PyXList_Type)
#define PyXList_CheckExact(op) ((op)->ob_type == &PyXList_Type)
#define PyXList_GetTag(xlist) ((PyXListObject*)xlist)->tpinfo->tag
#define PyXList_GetUri(xlist) ((PyXListObject*)xlist)->tpinfo->uri
#define PyXList_GetPrefix(xlist) ((PyXListObject*)xlist)->tpinfo->prefix

#ifdef __cplusplus
}
#endif
#endif /* !Py_LISTOBJECT_H */
