import httplib
import urlparse
import satine
from satine.schemas import soap

soap_fault = satine.xml2py('<soap:Fault/>')
soap_envelope = satine.xml2py('<soap:Envelope/>')
soap_body = satine.xml2py('<soap:Body/>')
soap_header = satine.xml2py('<soap:Header/>')

soap_template = """<?xml version="1.0" encoding="UTF-8"?>
<soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">
  <soap:Body>
    %s
  </soap:Body>
</soap:Envelope>"""

class soap:
    def __init__(self, url, action):
        self.action = action
        self.protocol, self.host, self.path, self.params, self.query, self.fragment = urlparse.urlparse(url)
        self.responses = []
        self.headers = []

    def addHeader(self, header):
        self.headers.append(header)

    def emptyHeaders(self):
        self.headers = []
        
    def write(self, xlist):
        headers = {'Content-type':'text/xml', 'SOAPAction':self.action}
        content = satine.py2xml(xlist)
        request = soap_template % (content,)
        print request
        connection = httplib.HTTPConnection(self.host)
        connection.request("POST", self.path, request, headers)
        self.responses.append(connection.getresponse())

    def writeRaw(self, data):
        headers = {'Content-type':'text/xml', 'SOAPAction':self.action}
        request = soap_template % (data,)
        print request
        connection = httplib.HTTPConnection(self.host)
        connection.request("POST", self.path, request, headers)
        self.responses.append(connection.getresponse())
        
    def read(self):
        if not self.responses:
            return None
        response = self.responses.pop()
        xml = response.read()
        env = satine.xml2py(xml)
        py = env.query("<soap:Envelope><soap:Body>.|")[0]
        return py



class soapRPCMethod:
    soapBindings = {
        int:    'xsd:int',
        float: 'xsd:double',
        str:    'xsd:string',
        unicode:'xsd:string'
    }
    
    def __init__(self, ws, method):
        self.ws = ws
        self.method = method

    def soaptype(self, var):
        if not self.soapBindings.has_key(var.__class__):
            raise "unsupported type %s " (var.__class__,)
        return self.soapBindings[var.__class__]
                    
    def __call__(self, *args, **keywds):
        content = '<gs:%s xmlns:gs="%s">\n' % (self.method, self.ws.service)
        count = 0
        for varvalue in args:
            varname = "arg%i" % (count,)
            count += 1
            if self.ws.useXMLDataType:
                content += '<%s xsi:type="%s">%s</%s>\n' % (varname, self.soaptype(varname), varvalue, varname)
            else:
                content += '<%s>%s</%s>\n' % (varname, varvalue, varname)

        for varname in keywds:
            varvalue = keywds[varname]
            if self.ws.useXMLDataType:
                content += '<%s xsi:type="%s">%s</%s>\n' % (varname, self.soaptype(varname), varvalue, varname)
            elif varvalue == '':
                content += '<%s/>\n' % (varname,)
            else:
                content += '<%s>%s</%s>\n' % (varname, varvalue, varname)
                
        content += '</gs:%s>' % (self.method,)
        self.ws.writeRaw(content)
        res = self.ws.read()
        print res
     
class soapRPC(soap):
    def __init__(self, url, service, soapAction = "", useXMLDataType = 0):
        soap.__init__(self, url, soapAction)
        self.service = service
        self.useXMLDataType = useXMLDataType

    def __getattr__(self, name):
        return soapRPCMethod(self, name)
                
def test():
    google =  """
    <gs:doGoogleSearch xmlns:gs="urn:GoogleSearch">
      <restrict/>
      <filter>true</filter>
      <lr/>
      <q></q>
      <key></key>
      <start>0</start>
      <maxResults>10</maxResults>
      <safeSearch>false</safeSearch>
      <oe>latin1</oe>
      <ie>latin1</ie>
    </gs:doGoogleSearch>
    """
    py = satine.xml2py(google)
    py[4].append("00000000000000000000000000000000")
    py[3].append("absurd obfuscation")

    con = soap("http://api.google.com/search/beta2","urn:GoogleSearchAction")
    con.write(py)
    print con.read()
    rpc = soapRPC("http://api.google.com/search/beta2","urn:GoogleSearch")
    rpc.doGoogleSearch(key="00000000000000000000000000000000", q="absurd obfuscation",
                           start=0, maxResults=10, filter="true", restrict="",
                           safeSearch="false", lr="", ie="latin1", oe="latin1")

if __name__ == '__main__':
    test()