import satine
import unittest
from satine.stream import reader, writer
import StringIO
import os

def fullname(filename):
    testpath = satine.test.__path__[0]
    return testpath+os.sep+filename

def openfile(filename):
    testpath = satine.test.__path__[0]
    filename = testpath+os.sep+filename
    return open(filename, "rb")    

class Read(unittest.TestCase):
    def setUp(self):
        self.st = reader(openfile('addressbook1.xml'))
        
    def testRead(self):
        it = iter(self.st)
        item = it.next()
        self.assertEqual(item.tpinfo.tag, "addressbook")
        item = it.next()
        self.assertEqual(item.tpinfo.tag, "person")
        self.assertEqual(item.surname, "Gates")
        self.assertEqual(item["d:prova"], "ni")
        item = it.next()
        self.assertEqual(item, "The Microsoft chief")
        item = it.next()
        self.assertEqual(item, None)
        item = it.next()
        self.assertEqual(item.surname, "Tolvald")
        self.assertEqual(item.name, "Linus")
        item = it.next()
        self.assertEqual(item, "The Linux creator")
        item = it.next()
        self.assertEqual(item, None)
        item = it.next()
        self.assertEqual(item.surname, "Goethe")
        self.assertEqual(item.name, "Wolfgang")
        item = it.next()
        self.assertEqual(item, "German romantic poet")
        item = it.next()
        self.assertEqual(item, None)
        item = it.next()
        self.assertEqual(item.surname, "Eco")
        self.assertEqual(item.name, "Umberto")
        item = it.next()
        self.assertEqual(item, "Italian writer")
        item = it.next()
        self.assertEqual(item, None)
        item = it.next()
        self.assertEqual(item.surname, "Monet")
        self.assertEqual(item.name, "Claude")
        item = it.next()
        self.assertEqual(item, "French painter")
        item = it.next()
        self.assertEqual(item, None)
        item = it.next()
        self.assertEqual(item, None)
        self.assertRaises(StopIteration, it.next)        


    def testNone(self):
        noneCounter = 0
        for item in self.st:
            if item is None:
                noneCounter += 1
        self.assertEqual(noneCounter, 6)


class Write(unittest.TestCase):
    def setUp(self):
        self.st = reader(openfile('addressbook1.xml'))
        
    def testRead(self):        
        wst = StringIO.StringIO()
        out = writer(wst)
        for item in self.st:
            out.write([item])
        wst.seek(0)
        res = wst.read()
        sum = 0
        for c in res:
            sum += ord(c)
        self.assertEqual(len(res), 377)
        self.assertEqual(sum, 33365)

if __name__ == "__main__":
    unittest.main()        